"""Add custom puppet support

Revision ID: 26341fb32054
Revises: 3c5af010538a
Create Date: 2019-05-02 18:48:45.532175

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '26341fb32054'
down_revision = '3c5af010538a'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('puppet', sa.Column('access_token', sa.Text(), nullable=True))
    op.add_column('puppet', sa.Column('custom_mxid', sa.String(length=255), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('puppet') as batch_op:
        batch_op.drop_column('custom_mxid')
        batch_op.drop_column('access_token')
    # ### end Alembic commands ###
