# Generated in setup.py

git_tag = None
git_revision = 'e5c77061'
version = '0.1.0.dev16'
linkified_version = '0.1.0.dev16+[e5c77061](https://github.com/tulir/mautrix-facebook/commit/e5c770611d523b6db83ecb11d4aee46e608cc575)'
